#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.

"""
Classes and methods for laying the data location foundation for various types of CASA processing
"""
import json
import logging
import os
import shutil
import sys
import tarfile
from distutils.dir_util import copy_tree
from pathlib import Path
from posixpath import isfile
from typing import Optional

from casa_envoy.interfaces import FoundationIF
from casa_envoy.schema import AbstractTextFile


def set_permissions(path: Path):
    """
    reusable function for setting file permissions

    :param path: path to set permissions on
    :return:
    """
    for root, dirs, files in os.walk(path):
        for d in dirs:
            os.chmod(os.path.join(root, d), 0o755)
        for f in files:
            os.chmod(os.path.join(root, f), 0o755)


class GeneralFoundation(FoundationIF):
    """
    Class to set up the general data foundation for anything running CASA.
    - Special attention to user provided file locations, must be placed in working directory.
    """

    def __init__(self, parameters: dict, metadata: AbstractTextFile):
        self.logger = logging.getLogger("casa_envoy")
        self.parent_path = parameters["parent_path"]
        self.working_dir = self.parent_path / "working"
        self.content = json.loads(metadata.content)

    def data_foundation(self):
        """
        All user provided files should be placed in the 'working' directory inorder for CASA to pick them up.

        :return:
        """
        self.logger.info("Laying general data foundation...")

        if "providedFiles" in self.content:
            provided_files = self.content["providedFiles"].replace("[", "").replace("'", "").replace("]", "").split(",")

            # move provided files from parent directory into working directory for CASA
            for file in provided_files:
                stripped_file = file.strip()
                if stripped_file:
                    shutil.copy(stripped_file, self.working_dir)

        self.logger.info("General data foundation complete!")


class RestoreFoundation(FoundationIF):
    """
    Class to set up the data foundation for restores
    """

    def __init__(self, parameters: dict, metadata: AbstractTextFile):
        self.logger = logging.getLogger("casa_envoy")
        self.parameters = parameters
        self.parent_path = parameters["parent_path"]
        self.content = json.loads(metadata.content)
        self._piperestorescript_path: Optional[Path] = None

    def data_foundation(self):
        """
        CMS Restore requires two inputs: An EB and a Calibration for that EB
        After download, all data is in rawdata and the calibrations tables are contained
        in a tar file. We need to extract all the calibration tables to the /products directory
        for CASA processing.

        :return:
        """

        self.logger.info("Laying restore data foundation...")
        # ensure we're starting from the parent directory
        os.chdir(self.parent_path)
        self.extract_cal()
        path = Path("./products")
        self.logger.info("Ensuring correct file permissions....")
        set_permissions(path)
        CASAPIPERESTORESCRIPT = "casa_piperestorescript.py"
        self.logger.info(f"Checking for {CASAPIPERESTORESCRIPT})...")
        if (path / CASAPIPERESTORESCRIPT).is_file():
            dest = Path(f"working/{CASAPIPERESTORESCRIPT}")
            shutil.copyfile(path / CASAPIPERESTORESCRIPT, dest)
            self._piperestorescript_path = dest.resolve()
            self.logger.info(f"Found {CASAPIPERESTORESCRIPT}, copied it to {dest}")
        self.logger.info("Restore data foundation complete!")

    @property
    def piperestorescript_path(self) -> Optional[Path]:
        """Absolute path to the casapiperestorescript.py for this restore, or None if one doesn't exist"""
        return self._piperestorescript_path

    def extract_cal(self):
        self.logger.info("Extracting calibration tar file to products directory...")
        cal_name = self.content["fileSetIds"][1]
        cal_path = "./rawdata/" + cal_name + "/" + cal_name
        if Path(cal_path).exists():
            calibration = tarfile.open(cal_path)
            # extract to products
            calibration.extractall("./rawdata")
            calibration.close()
            src = "./rawdata/products"
            src2 = "./rawdata/" + cal_name[:-4] + "/products"
            if Path(src).exists():
                copy_tree(src, "./products/")
            elif Path(src2).exists():
                copy_tree(src2, "./products")
            else:
                self.logger.error(f"ERROR: calibration tar file {cal_name} failed to extract correctly!")

        else:
            self.logger.error(f"ERROR: calibration tar file {cal_name} not found in rawdata!")
            # abort since required data not found
            sys.exit(1)


class CubeFoundation(FoundationIF):
    """
    Class to set up the data foundation for VLASS Cubes
    """

    def __init__(self, parameters: dict):
        self.logger = logging.getLogger("casa_envoy")
        self.parameters = parameters
        self.parent_path = self.parameters["parent_path"]
        self.working_dir = self.parent_path / "working"
        self.products_dir = self.parent_path / "products"

    def data_foundation(self):
        """
        VLASS Cubes require two inputs: the reimaging resources of an SECI and the associated SEIP_parameters.list
        After download, all data is in rawdata and the reimaging resources are contained
        in a tar file. We need to extract all the reimaging resources from the tar and move the parameters.list file
        to the working directory for CASA processing.

        :return:
        """
        self.logger.info("Laying cube data foundation...")
        os.chdir(self.parent_path)
        # self.extract_reimaging_resources()
        self.move_ancillaries()

        self.logger.info("Ensuring correct file permissions....")
        set_permissions(self.working_dir)
        self.logger.info("Cube data foundation complete!")

    def extract_reimaging_resources(self):
        """
        Extract resources to working directory

        :return:
        """
        self.logger.info("Extracting reimaging resources to working directory...")
        resource_path = self.parent_path / "rawdata" / "reimaging_resources.tgz"
        if Path(resource_path).exists():
            resources = tarfile.open(resource_path)
            # extract to working
            resources.extractall(self.working_dir)
            resources.close()
        else:
            self.logger.error("ERROR: reimaging_resources.tgz not found in rawdata!")

    def move_ancillaries(self):
        """
        Move ancillary products to correct locations for processing

        :return:
        """
        list_file_path = Path("./rawdata/SEIP_parameter.list")
        resources_path = Path("./rawdata/reimaging_resources.tgz")
        component_path = list(Path("./rawdata").glob("*catalog.csv"))[0]

        if Path(list_file_path).exists():
            shutil.copy(list_file_path, self.working_dir)
        else:
            self.logger.error("ERROR: Could not find SEIP_parameter.list in rawdata!")

        if Path(resources_path).exists():
            shutil.copy(resources_path, self.working_dir)
        else:
            self.logger.error("ERROR: Could not find reimaging_resources.tgz in rawdata!")

        if Path(component_path).exists():
            shutil.copy(component_path, self.products_dir)
        else:
            self.logger.error("ERROR: Could not find component list in rawdata!")
