#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
import glob
import json
import logging
import os
import pathlib
import re
import subprocess
import sys
from typing import Dict, Union

from casa_envoy.auditor import AuditDirectories, AuditFiles
from casa_envoy.configure import (
    CasaConfigure,
    ParallelRunner,
    PiperestorescriptRunner,
    StandardRunner,
)
from casa_envoy.enums import ProductType
from casa_envoy.foundation import CubeFoundation, GeneralFoundation, RestoreFoundation
from casa_envoy.interfaces import LauncherIF, RunnerIF
from casa_envoy.schema import AbstractTextFile


def get_abs_file(filename: str) -> AbstractTextFile:
    with open(filename, "r") as file:
        content = file.read()
        file.close()
    return AbstractTextFile(filename, content)


def run_audit(logger: logging.Logger, ppr: AbstractTextFile, metadata: AbstractTextFile, parameters: Dict[str, str]):
    dir_audit = AuditDirectories([ppr, metadata], parameters).audit()
    if dir_audit:
        logger.info("Directory audit successful!")
    else:
        logger.error("FAILURE: directory structure audit was unsuccessful!")
        sys.exit(1)

    audit = AuditFiles([ppr, metadata], parameters).audit()
    if audit:
        logger.info("File audit successful!")
    else:
        logger.error("FAILURE: file audit was unsuccessful!")
        sys.exit(1)


def determine_runner_type(parameters: dict) -> RunnerIF:
    """
    determine the type of CASA runner needed: Standard or Parallel

    :param parameters: dictionary of parameters
    :return: typed Runner
    """
    if "requested_parallel" in parameters:
        return ParallelRunner(parameters)
    elif "piperestorescript_path" in parameters:
        # This function is currently called when the Launcher is initialized,
        #     i.e. before the RestoreFoundation would know if there's a piperestorescript or not,
        #     so its results will be overriden if the RestoreFoundation finds a piperestorescript.
        # This case is here only in case that order of operations changes.
        return PiperestorescriptRunner(parameters)
    else:
        return StandardRunner(parameters)


class CasaLauncher:
    def __init__(self, runner: RunnerIF):
        self.logger = logging.getLogger("casa_envoy")
        self.runner = runner
        self.parameters = runner.parameters
        self.metadata = get_abs_file(self.runner.parameters.get("metadata"))

    def run(self):
        """
        Run CASA

        :return:
        """
        CasaConfigure(self.parameters, json.loads(self.metadata.content)).setup_environment()

        # Check if running locally or with rapid testing parameters, if so, run Vela instead
        if self.parameters.get("useCasa"):
            self.runner.call_casa()
            self.check_logs()
        else:
            self.logger.info("RUNNING VELA!")
            run_type = self.parameters.get("product_type")
            metadata = self.parameters.get("metadata")
            ppr = self.parameters.get("ppr")
            # casa_envoy might be running in a condor scratch directory or
            # directly from the processing directory. Account for both cases
            vela_path = (
                "./vela"
                if (pathlib.Path.cwd() / "vela").exists()
                else f"/lustre/aoc/cluster/pipeline/{os.environ.get('CAPO_PROFILE')}/workspaces/sbin/vela"
            )
            subprocess.run([vela_path, "-f", metadata, ppr, "--" + run_type])

    def check_logs(self):
        """
        Check CASA log for 'SEVERE' errors or 'Traceback' logs and fail if necessary

        :return:
        """
        self.logger.info("CHECKING CASA LOGS!")
        # make sure we are in the correct directory to find log file
        if not os.getcwd().endswith("/working"):
            os.chdir(self.parameters.get("parent_path") / "working")

        casa_logs = glob.glob("casa-*.log")

        for file in casa_logs:
            if re.match("\bSEVERE\b", open(file).read()):
                self.logger.error("CASA ERROR!")
                sys.exit(1)
            elif re.match("\bTraceback\b", open(file).read()):
                self.logger.error("CASA ERROR!")
                sys.exit(1)
            else:
                self.logger.info("CASA Success!")


class CalibrationLauncher(LauncherIF):
    def __init__(self, parameters: dict):
        self.logger = logging.getLogger("casa_envoy")
        self.parameters = parameters
        self.ppr = get_abs_file(parameters.get("ppr"))
        self.metadata = get_abs_file(parameters.get("metadata"))
        self.content = json.loads(self.metadata.content)
        self.runner = determine_runner_type(parameters)

    def launch_casa(self):
        check_input = (
            self.check_restorable()
            if self.parameters["product_type"] == ProductType.RESTORE.value
            else self.check_calibratable()
        )

        if check_input:
            self.prepare_for_casa()
            run_audit(self.logger, self.ppr, self.metadata, self.parameters)
            CasaLauncher(self.runner).run()
        else:
            self.logger.error("ERROR: Provided SPL(s) are not correct type(s) for processing!")
            sys.exit(1)

    def prepare_for_casa(self):
        # Ensure all data is in the required locations for CASA processing (This is not always rawdata!)
        GeneralFoundation(self.parameters, self.metadata).data_foundation()
        if self.parameters["product_type"] == ProductType.RESTORE.value:
            restore_foundation = RestoreFoundation(self.parameters, self.metadata)
            restore_foundation.data_foundation()
            if restore_foundation.piperestorescript_path:
                # Need to make this available to the Runner if it exists
                self.parameters["piperestorescript_path"] = str(restore_foundation.piperestorescript_path)
                # Since determine_runner_type() is called before this, need to override it if a piperestorescript is found
                self.runner = PiperestorescriptRunner(self.parameters)
        return

    def check_calibratable(self) -> bool:
        spl = self.content["productLocator"]
        if "execblock" in spl:
            return True
        else:
            self.logger.error("SPL ERROR: This product locator is not calibratable! Please check your inputs")
            return False

    def check_restorable(self) -> bool:
        spl = self.content["productLocator"]
        if "calProductLocator" in self.content:
            cal_spl = self.content["calProductLocator"]
            if "execblock" in spl and "calibration" in cal_spl:
                return True

        self.logger.error("SPL ERROR: This set of product locators are not restorable! Please check your inputs")
        return False


class ImagingLauncher(LauncherIF):
    def __init__(self, parameters: dict):
        self.logger = logging.getLogger("casa_envoy")
        self.parameters = parameters
        self.ppr = get_abs_file(parameters.get("ppr"))
        self.metadata = get_abs_file(parameters.get("metadata"))
        self.content = json.loads(self.metadata.content)
        self.runner = determine_runner_type(parameters)

    def launch_casa(self):
        check_input = (
            self.check_cal_and_imageable()
            if self.parameters["product_type"] == ProductType.INTEGRATED_CAL_IMG.value
            else self.check_imageable()
        )

        if check_input:
            self.prepare_for_casa()
            run_audit(self.logger, self.ppr, self.metadata, self.parameters)
            CasaLauncher(self.runner).run()
        else:
            self.logger.error("ERROR: CMS information missing or incorrect!")
            sys.exit(1)

    def prepare_for_casa(self):
        wf_name = self.content["workflowName"]
        GeneralFoundation(self.parameters, self.metadata).data_foundation()

        if "restore" in wf_name:
            RestoreFoundation(self.parameters, self.metadata).data_foundation()
        if wf_name == "vlass_coarse":
            CubeFoundation(self.parameters).data_foundation()

        return

    def check_imageable(self) -> bool:
        if self.parameters["product_type"] == ProductType.VLASS_IMG.value:
            # We're trusting the VLASS manager to know what it's doing (dangerous I know)
            return True

        cms_name = self.content["cmsName"]
        cms_path = self.content["calibrationSourceDirectory"]
        if cms_name is not None and cms_path is not None and cms_name[-3:] == ".ms":
            return True
        else:
            self.logger.error("CMS ERROR: Imaging requires a valid CMS name and location!")
            return False

    def check_cal_and_imageable(self) -> bool:
        wf_name = self.content["workflowName"]
        spl = self.content["productLocator"]

        if "restore" in wf_name:
            cal_spl = self.content["calProductLocator"]

            if "execblock" in spl and "calibration" in cal_spl:
                return True
            else:
                self.logger.error(
                    "SPL ERROR: This set of product locators are not restorable! Please check your inputs"
                )
        elif "cal" in wf_name:
            if "execblock" in spl:
                return True
            else:
                self.logger.error("SPL ERROR: This product locator is not calibratable! Please check your inputs")
                return False


class SplitLauncher(LauncherIF):
    def __init__(self, parameters: dict):
        self.logger = logging.getLogger("casa_envoy")
        self.parameters = parameters
        self.ppr = get_abs_file(parameters.get("ppr"))
        self.metadata = get_abs_file(parameters.get("metadata"))
        self.content = json.loads(self.metadata.content)
        self.runner = determine_runner_type(parameters)

    def launch_casa(self):
        self.prepare_for_casa()
        run_audit(self.logger, self.ppr, self.metadata, self.parameters)
        CasaLauncher(self.runner).run()

    def prepare_for_casa(self):
        GeneralFoundation(self.parameters, self.metadata).data_foundation()
